/****************************************************************************************/
/*											*/
/* This program is free software; you can redistribute it and/or modify it under the	*/
/* terms of the GNU General Public License as published by the Free Software		*/
/* Foundation; either version 2 of the License, or (at your option) any later version.	*/
/*											*/
/* This program is distributed in the hope that it will be useful, but WITHOUT ANY	*/
/* WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A	*/
/* PARTICULAR PURPOSE. See the GNU General Public License for more details.		*/
/*											*/
/* You should have received a copy of the GNU General Public License along with this	*/
/* program; (See "LICENSE.GPL"). If not, write to the Free Software Foundation, Inc.,	*/
/* 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.				*/
/*											*/
/*--------------------------------------------------------------------------------------*/
/*											*/
/*		Joerg Anders, TU Chemnitz, Fakultaet fuer Informatik, GERMANY		*/
/*		ja@informatik.tu-chemnitz.de						*/
/*											*/
/*											*/
/****************************************************************************************/

#include <stdlib.h> //  esigra: needed for mkstemp when --without-libs
#include <stdio.h>
#include <kmessagebox.h>
#include <klocale.h>
#include <qstring.h>
#include <qregexp.h>
#include "filehandler.h"
#include "resource.h"
#include "clef.h"
#include "timesig.h"
#include "sign.h"
#include "rest.h"
#include "chord.h"
#include "keysig.h"
#include "mainframewidget.h"
#include "voice.h"
#include "outputbox.h"
#include "parsertypes.h"
#include "staff.h"
#include "uiconnect.h"

NFileHandler::NFileHandler() : newLines_("\n") {
	int i;
	os_ = new ostrstream(buffer_, 128);
	ornaments_ = new ostrstream(ornamentbuffer_,  128);
	dynamics_  = new ostrstream(dynamicsbuffer_, 128);
	for (i = 0; i < NUM_LYRICS; i++) {
		lyricsLine_[i] = new ostrstream(lyricsbuffer_[i], 128);
	}
	badlist_.setAutoDelete(true);
	fatallist_.setAutoDelete(true);
	mupWarn_ = new mupWrn( 0 );
}

void NFileHandler::pitchOut( const NNote *note, NClef *ac_clef, bool with_tie) {
	int i, octave;

	out_ << ac_clef->line2Name(note->line, &octave, false);
	if (!(note->status & STAT_PART_OF_TIE) && with_tie) {
		if (note->status & STAT_FORCE) {
			switch (note->offs) {
				case  1: out_ << "#"; break;
				case -1: out_ << "&"; break;
				case  0: out_ << "n"; break;
				case  2: out_ << "x"; break;
				case -2: out_ << "&&"; break;
			}
		}
		else {
			switch (note->status & ACC_MASK) {
				case STAT_CROSS: out_ << "#";  break;
				case STAT_FLAT: out_ << "&";   break;
				case STAT_NATUR: out_ << "n";  break;
				case STAT_DCROSS: out_ << "x"; break;
				case STAT_DFLAT: out_ << "&&"; break;
			}
		}
	}
	if (octave > 0) {
		for (i = 0; i < octave; i++) {
			out_ << "+";
		}
	}
	else if (octave < 0) {
		for (i = 0; i > octave; i--) {
			out_ << "-";
		}
	}
	if ((note->status & STAT_TIED) && with_tie) out_ << "~";
	switch (note->status & BODY_MASK) {
		case STAT_BODY_CROSS: out_ << " bcr"; break;
		case STAT_BODY_CROSS2: out_ << " bcr2"; break;
		case STAT_BODY_CIRCLE_CROSS: out_ << " bcrc"; break;
		case STAT_BODY_TRIA: out_ << " btr"; break;
		case STAT_BODY_RECT: out_ << " brec"; break;
	}
}


	
bool NFileHandler::writeStaffs(QString fname, QList<NStaff> *stafflist, NMainFrameWidget *mainWidget, bool showErrors) {
	NVoice *voice_elem;
	NStaff *staff_elem;
	NMusElement *last_elem;
	int i, j, written_voices;
	int multirestlength;
	int current_time;
	int voice_count;
	NTimeSig *timesig;
	bar_nr_ = 2;
	int btype, bartype;
	bool first;
	bool ending2seen = false;
	badmeasure *bad;
	QString outstring;
	
	countof128th_ = 128;
	drum_problem_written_ = false;

	badlist_.clear();
	fatallist_.clear();

	staffCount_ = stafflist->count();
	pending_elems_ = new pending_elements[staffCount_];
	pending_multi_rests_ = new int*[staffCount_];
	for (i = 0, staff_elem = stafflist->first(); staff_elem; staff_elem = stafflist->next(), i++) {
		pending_elems_[i].pending_program_change = 
		pending_elems_[i].pending_volsig = 0;
		pending_multi_rests_[i] = new int[staff_elem->voiceCount()];
		for (j = 0; j < staff_elem->voiceCount(); j++)  {
			*(pending_multi_rests_[i] + j) = 0;
		}
	}
	pending_tempo_sig_ = 0;
		

	if (stafflist->isEmpty()) {
		KMessageBox::sorry
		  (0,
		   i18n("nothing to write"),
		   kapp->makeStdCaption(i18n("Write Staffs"))
		  );
		return false;
	}

	out_.open(fname);
	
	if (!out_) {
		*os_ << i18n("error opening file ") << fname << '\0';
		KMessageBox::sorry
		  (0, QString(os_->str()), kapp->makeStdCaption(i18n("Write Staffs")));
		return false;
	}

	out_ << "header" << endl;
	if (!mainWidget->scTitle_.isEmpty()) {
		outstring = mainWidget->scTitle_;
		outstring.replace (newLines_, "\\n");
		out_ << "\ttitle bold (" <<  SC_TITLE_FONT_SIZE << ") \"" << outstring << '"' << endl;
	}
	if (!mainWidget->scSubtitle_.isEmpty()) {
		outstring = mainWidget->scSubtitle_;
		outstring.replace (newLines_, "\\n");
		out_ << "\ttitle bold (" << SC_SUBTITLE_FONT_SIZE << ") \"" << outstring << '"' << endl;
	}
	if (!mainWidget->scAuthor_.isEmpty()) {
		outstring = mainWidget->scAuthor_;
		outstring.replace (newLines_, "\\n");
		out_ << "\ttitle bold (" << SC_AUTHOR_FONT_SIZE << ") \"" << outstring << '"' << endl;
	}
	if (!mainWidget->scLastAuthor_.isEmpty()) {
		outstring = mainWidget->scLastAuthor_;
		outstring.replace (newLines_, "\\n");
		out_ << "\ttitle bold (" << SC_LAST_AUTHOR_FONT_SIZE << ") \"" << outstring << '"' << endl;
	}
	if (!mainWidget->scComment_.isEmpty()) {
		outstring = mainWidget->scComment_;
		outstring.replace (newLines_, "\\n");
		out_ << "\ttitle bold (" << SC_COMMENT_FONT_SIZE << ") \"" <<  outstring << '"' << endl;
	}
	out_ << "score" << endl;
	if (mainWidget->paramsEnabled()) {
		out_ << "\tunits = cm" << endl;
		out_ << "\tpagewidth = " << ((double) mainWidget->getSaveWidth() / 10.0) << endl;
		out_ << "\tpageheight = " << ((double) mainWidget->getSaveHeight() / 10.0) << endl;
	}
	if (mainWidget->withMeasureNums()) {
		out_ << "\tmeasnum = y" << endl;
	}
	out_ << "\tstaffs = " << stafflist->count() << endl;
	voice_elem = stafflist->first()->getVoiceNr(0);
	timesig = voice_elem->getFirstTimeSig();
	if (timesig) {
		countof128th_ = timesig->numOf128th();
		out_ << "\ttime = " << timesig->getNumerator() <<
			'/' << timesig->getDenominator() << endl;
		curr_num_ = timesig->getNumerator();
		curr_denom_ = timesig->getDenominator();
	}
	else {
		countof128th_ = 128;
		curr_num_ = curr_denom_ = 4;
	}
	for (staff_elem = stafflist->first(); staff_elem; staff_elem = stafflist->next()) {
		voice_count = staff_elem->voiceCount();
		for (j = 0; j < voice_count; j++) {
			voice_elem = staff_elem->getVoiceNr(j);
			voice_elem->prepareForWriting();
		}
	}
	
	first = true;
	musicmode_ = true;
	do {
		written_voices = 0;
		for (i = 1, staff_elem = stafflist->first(); staff_elem; staff_elem = stafflist->next(), i++) {
			voice_elem = staff_elem->getVoiceNr(0);
			hasClef_ = false;
			writeScoreInfo(i, voice_elem, first);
			if (first && !hasClef_) {
				*os_
				  << i18n("Please choose a clef for staff! ") << i
				  << i18n(" Otherwise this file will be unreadable in future.") << '\0';
				bad = new badmeasure(ERR_NO_CLEF, i, bar_nr_, 3 /* dummy */, countof128th_);
				fatallist_.append(bad);
			}
		}
 		if (!musicmode_) {
			musicmode_ = true;
			out_ << "music" << endl;
		}
		if (first) {
			for (i = 1, staff_elem = stafflist->first(); staff_elem; staff_elem = stafflist->next(), i++) {
				out_ << "midi " << i << " 1: 0 \"channel=" << staff_elem->getChannel()+1 << 
				"\"; 0 \"program=" << staff_elem->getVoice() << "\";" << endl;
			}
			for (i = 1, staff_elem = stafflist->first(); staff_elem; staff_elem = stafflist->next(), i++) {
				out_ << "midi " << i << " 1: 0 \"parameter=" << PROG_VOL << "," << staff_elem->getVolume() << "\";" << endl;
				out_ << "midi " << i << " 1: 0 \"parameter=" << PROG_CHORUS << "," << staff_elem->chorus_ << "\";" << endl;
				out_ << "midi " << i << " 1: 0 \"parameter=" << PROG_REVERB << "," << staff_elem->reverb_ << "\";" << endl;
				out_ << "midi " << i << " 1: 0 \"parameter=" << PROG_PAN << "," << staff_elem->pan_ << "\";" << endl;
			}
		}
		multirestlength = determineMultiRest(stafflist);
		bartype = 0;
		for (i = 1, staff_elem = stafflist->first(); staff_elem; staff_elem = stafflist->next(), i++) {
			voice_elem = staff_elem->getVoiceNr(0);
			ornaments_->seekp(0);
			dynamics_->seekp(0);
			btype = writeStaffUntilBar(i, voice_elem, first, multirestlength);
			voice_count = staff_elem->voiceCount();
			last_elem = voice_elem->getCurrentPosition();
			if (last_elem) {
				current_time = last_elem->midiTime_;
			}
			else {
				current_time = (1 << 30);
			}
			for (j = 2; j <= voice_count; j++) {
				voice_elem = staff_elem->getVoiceNr(j-1);
				writeVoiceElemsTill(i, j, voice_elem, current_time, multirestlength);
			}
			if (ornaments_->pcount() > 1) {
				*ornaments_ << endl << '\0';
				out_ << ornaments_->str();
			}
			if (dynamics_->pcount() > 1) {
				*dynamics_ << endl << '\0';
				out_ << dynamics_->str();
			}
			if (btype) {
				if (bartype) {
					if (btype != bartype) {
						bad = new badmeasure(ERR_BARTYPES, i, bar_nr_, 3 /* dummy */, countof128th_);
						fatallist_.append(bad);
					}
				}
				bartype = btype;
				written_voices++;
			}
		}
		if (written_voices) {
			if (!bartype) bartype = SIMPLE_BAR;
			switch (bartype & BAR_SYMS) {
				case SIMPLE_BAR: out_ << "bar"; break;
				case REPEAT_OPEN: out_ << "repeatstart"; break;
				case REPEAT_CLOSE: out_ << "repeatend"; break;
				case REPEAT_OPEN_CLOSE: out_ << "repeatboth"; break;
			}
			if (ending2seen) {
				ending2seen = false;
				out_ << " endending";
			}
			switch (bartype & SPECIAL_ENDING) {
				case SPECIAL_ENDING1: out_ << " ending \"1.\""; break;
				case SPECIAL_ENDING2: out_ << " ending \"2.\""; ending2seen = true; break;
			}

			out_ << " // " << bar_nr_++ << endl;
		}
		first = false;
	}
	while (written_voices);
	out_.close();
	delete pending_elems_;
	for (i = 0, staff_elem = stafflist->first(); staff_elem; staff_elem = stafflist->next(), i++) {
		delete pending_multi_rests_[i];
	}
	delete pending_multi_rests_;

	if (!showErrors) return true;

	if (!fatallist_.isEmpty()) {
		QString output;
		for (bad = fatallist_.first(); bad; bad = fatallist_.next()) {
			switch (bad->kind) {
				case ERR_BARTYPES:
					output += i18n
						("staff %1; measure %2: you mixed different bar types; this "
						 "cannot be expressed in MUP and causes data loss\n"
						).arg(bad->track).arg(bad->measure);
					break;
				case ERR_TEMPO_PLACEMENT:
					output += i18n
						("staff %1; measure %2: you placed a tempo signature not "
						 "immediately after a bar sign;\n\tthis cannot be expressed in "
						 "MUP and causes data loss\n"
						).arg(bad->track).arg(bad->measure);
					break;
				case ERR_VOL_PLACEMENT:
					output += i18n
						("staff %1; measure %2: you placed a volume signature not "
						 "immediately after a bar sign;\n\tthis cannot be expressed in "
						 "MUP and causes data loss\n"
						).arg(bad->track).arg(bad->measure);
					break;
				case ERR_PROG_PLACEMENT:
					output += i18n
						("staff %1; measure %2: you placed an instrument change not "
						 "immediately after a bar sign;\n\tthis cannot be expressed in "
						 "MUP and causes data loss\n"
						).arg(bad->track).arg(bad->measure);
					break;
				case ERR_NO_CLEF:
					output += i18n
						("Staff %1 has no clef. This causes data loss and can inhibit "
						 "correct restoring!\n"
						).arg(bad->track);
					break;
				case ERR_MULTIREST:
					output += i18n
						("Staff %1 ; measure %2 :The multirests are not aligned\n"
						).arg(bad->track).arg(bad->measure);
					break;
			}
		}
		OutputBox::warning(0, i18n("Possible data loss!"), output, i18n("Save"));
	}

	if (!NResource::dontShowMupWarnings_ && !badlist_.isEmpty()) {
		QString output;
		output = i18n
			("Noteedit will save your score, and it can (probably) restore the score.\n"
			 "But the file doesn't conform to MUP, so it will not work with the MUP interpreter.\n");
		output += i18n("-----------------------------------------------------\n");
		for (bad = badlist_.first(); bad; bad = badlist_.next()) {
			switch (bad->kind) {
				case ERR_NOTE_COUNT:
					output += i18n
						("staff %1; measure  %2 : %3 128th, should be: %4\n").arg(bad->track).
						arg(bad->measure).arg(bad->realcount).arg(bad->shouldbe);
					break;
				case ERR_LYRICS_COUNT:
					output += i18n(
						"staff %1; measure %2: Mup cannot deal with mixed lyrics/non-lyrics notes in the same measure\n").
						arg(bad->track).arg(bad->measure);
					break;
				case ERR_SLUR:
					output += i18n(
						"staff %1; measure %2: Mup up cannot deal with slurs over more than 1 element\n").
						arg(bad->track).arg(bad->measure);
					break;
				case ERR_LONELY_TRILL:
					output += i18n(
						"staff %1; measure %2: Mup cannot deal with trills without \"tr\" sign\n").
						arg(bad->track).arg(bad->measure);
					break;
				case ERR_IRREGULAER:
					output += i18n
						("Staff %1 has irregular key, this may not work with MUP.\n")
						.arg(bad->track);
					break;
				case ERR_TOO_MANY_VOICES:
					output += i18n
						("Staff %1 has too many voices! MUP accepts up to 3 voices.\n")
						.arg(bad->track);
					break;
				case ERR_DRUM_STAFF:
					output += i18n
						("Staff %1 contains drum staff. This cannot be expressed in MUP.\n")
						.arg(bad->track);
					break;
			}
		}
		mupWarn_->setOutput(&output);
		mupWarn_->show();
	}

	return true;
}

void NFileHandler::writeScoreInfo(int voice_nr, NVoice *voi, bool firstcall) {
	NMusElement *elem;
	NVoice *voice_elem;
	int voice_count, i;
	NClef *clef;
	NKeySig *ksig;
	NTimeSig *timesig;
	NStaff *actual_staff;
	int kind, count;
	bool first = true;

	actual_staff = voi->getStaff();
	elem = voi->getCurrentPosition();
	if (!elem) return;
	while(1) {
		switch(elem->getType()) {	
			case T_CLEF: 
				if (musicmode_) {
					musicmode_ = false;
					out_ << "score" << endl;
				}
				if (first) {
					out_ << "staff " << voice_nr << endl;
					if (firstcall) {
						voice_count = actual_staff->voiceCount();
						if (voice_count > 1) {
							out_ << "\tvscheme = " << voice_count << "o" << endl;
							if (voice_count > 3) {
								badmeasure *bad = new badmeasure(ERR_TOO_MANY_VOICES, voice_nr, bar_nr_, 3 /* dummy */, countof128th_);
								badlist_.append(bad);
							}
						}
						if (!actual_staff->staffName_.isEmpty()) {
							out_ << "label = \"" << actual_staff->staffName_ << '"' << endl;
						}
						out_ << "// overlength = " << actual_staff->overlength_ << endl;
						out_ << "// underlength = " << actual_staff->underlength_ << endl;
						out_ << "// lyricsdist = " << actual_staff->staff_props_.lyricsdist << endl;
						if (actual_staff->transpose_) {
							out_ << "// playtransposd = " << actual_staff->transpose_ << endl;
						}
						for (i = 0; i < voice_count; i++) {
							voice_elem = actual_staff->getVoiceNr(i);
							if (voice_elem->yRestOffs_ != 0) {
								out_ << "// yrestoffs " << (i + 1) << " = " << voice_elem->yRestOffs_ << endl;
							}
							switch(voice_elem->stemPolicy_) {
							case STEM_POL_UP: out_ << "// stempolicy " << (i + 1) << " = stemup" << endl;break;
							case STEM_POL_DOWN: out_ << "// stempolicy " << (i + 1) << " = stemdown" << endl;break;
							}
						}
					}
				}
				clef = (NClef *) elem;
				hasClef_ = true;
				actual_staff->actualClef_.change(clef);
				switch (clef->getSubType()) {
				case TREBLE_CLEF:
					out_ << "\tclef=treble" << endl;
					switch (clef->getShift()) {
				 		case -12: out_ << "\tdefoct=3" << endl; break; 
				   		case  12: out_ << "\tdefoct=5" << endl; break; 
					}
					break;
				case DRUM_CLEF:
					out_ << "\tclef=drum" << endl;
					if (!drum_problem_written_) {
						drum_problem_written_ = true;
						badlist_.append(new badmeasure(ERR_DRUM_STAFF, voice_nr, 0 /*dummy */, 3 /* dummy */, countof128th_));
					}
					break;
				case ALTO_CLEF:
					out_ << "\tclef=alto" << endl;
					switch (clef->getShift()) {
				 		case -12: out_ << "\tdefoct=3" << endl; break; 
				   		case  12: out_ << "\tdefoct=5" << endl; break; 
					}
					break;
				case TENOR_CLEF:
					out_ << "\tclef=tenor" << endl;
					switch (clef->getShift()) {
				 		case -12: out_ << "\tdefoct=3" << endl; break; 
				   		case  12: out_ << "\tdefoct=5" << endl; break; 
					}
					break;
				case BASS_CLEF:
					out_ << "\tclef=bass" << endl;
					switch (clef->getShift()) {
						case -12: out_ << "\tdefoct=2" << endl; break; 
						case  12: out_ << "\tdefoct=4" << endl; break; 
					}
					break;
				case DRUM_BASS_CLEF:
					out_ << "\tclef=drum_bass" << endl;
					if (!drum_problem_written_) {
						drum_problem_written_ = true;
						badlist_.append(new badmeasure(ERR_DRUM_STAFF, voice_nr, 0 /*dummy */, 3 /* dummy */, countof128th_));
					}
					break;
				default: hasClef_ = false;
				}
				break;
			case T_KEYSIG:
			        if (musicmode_) {
				   musicmode_ = false;
				   out_ << "score" << endl;
				}
				if (first) {
					out_ << "staff " << voice_nr << endl;
				}
				ksig = (NKeySig *) elem;
				if (ksig->isRegular(&kind, &count)) {
					out_ << "\tkey = " << count << ((kind == STAT_CROSS) ? "#" : "&") << endl;
				}
				else {
					badmeasure *bad = new badmeasure(ERR_IRREGULAER, voice_nr, bar_nr_, 3 /* dummy */, countof128th_);
					badlist_.append(bad);
					out_ << "//\tirregular = " << ksig->printKeys() << endl;
				}
				break;
			case T_TIMESIG:
				timesig = (NTimeSig *) elem;
				if (curr_num_ == timesig->getNumerator() && curr_denom_ == timesig->getDenominator()) break;
				musicmode_ = false;
				out_ << "score" << endl;
				out_ << "\ttime = " << timesig->getNumerator() <<
					'/' << timesig->getDenominator() << endl;
				countof128th_ = timesig->numOf128th();
				curr_num_ = timesig->getNumerator();
				curr_denom_ = timesig->getDenominator();
				break;
			default: return;
		}
		elem = voi->getNextPosition();
		if (!elem) return;
		first = false;
	}
}

int NFileHandler::writeStaffUntilBar(int staff_nr, NVoice *voi, bool first, int multirestlength) {
	NMusElement *elem;
	QString *lyrics;
	int total = 0;
	int part, tt;
	NNote *note, *note2;
	NChord *chord, *partner;
	NSign *sign;
	bool inbeam = false;
	bool intriplet = false;
	int subtype;
	int dest_measure_start_time;
	int count_of_measures;
	QString timestring;
	int i, dist;
	bool loop1 = true, loop2;
	int lyricsLineCount = 0;
	int measure_start_time = -1;
	double starttime, endtime;
	badmeasure *bad;
	NStaff *actual_staff;

	

	for (i = 0; i < NUM_LYRICS; i++) {
		lyricsLine_[i]->seekp(0);
	}
	actual_staff = voi->getStaff();
	elem = voi->getCurrentPosition();
	if (first) {
		while (elem && elem->getType() != T_CHORD && elem->getType() != T_REST) {
			switch (elem->getType()) {
				case T_SIGN:
		  			switch (elem->getSubType ()) {
						case REPEAT_OPEN:
						case REPEAT_CLOSE:
								subtype = elem->getSubType();
				 				elem = voi->getNextPosition();
				 				return subtype;
								break;
						case TEMPO_SIGNATURE:
								sign = (NSign *) elem;
								out_ << "midi all: 0 \"tempo=" << sign->getTempo() <<
									 "\";" << endl;
								out_ << "rom above all: 1 \"( \\(sm4n) = " << 
									sign->getTempo() << " )\";" << endl;
									pending_tempo_sig_ = 0;
								break;
					}
				 	elem = voi->getNextPosition();
					break;
				default: elem = voi->getNextPosition();
					break;
			}
		  }
		  for (i = 0; i < staffCount_; i++) {
			if (pending_elems_[i].pending_volsig) {
				writeVolSig(i + 1, pending_elems_[i].pending_volsig);
				pending_elems_[i].pending_volsig = 0;
			}
			if (pending_elems_[i].pending_program_change) {
				writeVolSig(i + 1, pending_elems_[i].pending_program_change);
				pending_elems_[i].pending_program_change = 0;
			}
		}
	
		  
	}
	if (elem == 0) return 0;
	if (multirestlength) {
		if (staff_nr == 1) {
			out_ << "multirest " << multirestlength << endl;
		}
		while(elem && elem->getType() != T_SIGN && !(elem->getSubType() & BAR_SYMS))  {
			elem = voi->getNextPosition();
		}
		loop1 = false;
		total = 3*countof128th_;
	}
	if (measure_start_time < 0) {
		measure_start_time = elem->midiTime_;
	}
	do {
		switch (elem->getType()) {
			case T_CHORD: if (loop1) {
					loop1 = false;
					out_ << staff_nr << " 1: ";
				     }
				     part = elem->getSubType(); 
				     total += (elem->status_ & STAT_TRIPLET) ? 2 * part / 3 : part;
				     switch (elem->status_ & DOT_MASK) {
					case 1: total += part / 2; break;
					case 2: total += 3 * part / 4; break;
				     }
				     if ((elem->status_ & STAT_TRIPLET) && !intriplet) {
					intriplet = true;
					out_ << "{ ";
				     }
				     if (total > 3*countof128th_) {
					total = 0;
					bad = new badmeasure(ERR_NOTE_COUNT, staff_nr, bar_nr_, total / 3, countof128th_);
					badlist_.append(bad);
				     }
				     chord = (NChord *) elem;

				     if (chord->status_ & STAT_SFORZ || chord->status_ & STAT_PORTA ||
				         chord->status_ & STAT_STPIZ || chord->status_ & STAT_SFZND ||
					 chord->status_ & STAT_STACC || chord->status_ & STAT_FERMT) {
				    	    out_ << "[with ";
				    	    if (chord->status_ & STAT_STACC)
						out_ << ".";
				    	    if (chord->status_ & STAT_SFORZ) 
					        out_ << "^";
					    if (chord->status_ & STAT_PORTA)
					        out_ << "-";
					    if (chord->status_ & STAT_STPIZ)
					        out_ << ",";
					    if (chord->status_ & STAT_SFZND)
					        out_ << ">";
					    if (chord->status_ & STAT_FERMT)
						out_ << "\"\\(ferm)\"";
					    out_ << "] ";
				     }
				     if (chord->getSubType() <= 3*64) {
					if (((chord->status_ & STAT_STEM_UP) &&  chord->getNoteList()->first()->line > 3)  ||
					    (!(chord->status_ & STAT_STEM_UP) && chord->getNoteList()->first()->line  < 4)) {
						out_ << ((chord->status_ & STAT_STEM_UP) ? "[up]" : "[down] ");
					}
				     }
				     if (part == 3*256) {
					out_ << "1/2";
				     }
				     else {
				     	out_ << 3*128 / part;
				     }
				     switch (elem->status_ & DOT_MASK) {
					case 1:	out_ << "."; break;
					case 2:	out_ << ".."; break;
				     }
				     loop2 = true;
			  	     for (note = elem->getNoteList()->first(); note; note = elem->getNoteList()->next()) {
					     pitchOut(note, &(actual_staff->actualClef_), true);
					     if (!drum_problem_written_ && (note->status & BODY_MASK)) {
						drum_problem_written_ = true;
						bad = new badmeasure(ERR_DRUM_STAFF, staff_nr, 0 /*dummy */, 3 /* dummy */, countof128th_);
						badlist_.append(bad);
					     }
					     if (loop2) {
						loop2 = false;
						if (chord->status_ & STAT_SLURED) {
							dist = voi->computeSlurDist(chord);
							if (dist > 1) {
								bad = new badmeasure(ERR_SLUR, staff_nr, bar_nr_, 0, 0);
								badlist_.append(bad);
								out_ << "<" << dist << ">";
							}
							else {
								partner = chord->getSlurPartner();
								note2 = partner->getNoteList()->first();
								out_ << "<" ;
					    			pitchOut( note2, &(actual_staff->actualClef_), false);
								out_ << ">" ;
							}
						}
					    }
				     }
				     if (elem->status_ & STAT_BEAMED) {
					if (!inbeam) {
						out_ << " bm";
						inbeam = true;
					 }
				     	 else if (elem->lastBeamed()) {
						out_ << " ebm";
						inbeam = false;
					 }
				     }
				     out_ << "; ";
				     if (elem->status_ & STAT_LAST_TRIPLET) {
					intriplet = false;
					out_ << " } above 3; ";
				     }
				     if (elem->trill_) {
					dest_measure_start_time = measure_start_time;
					tt = voi->findTimeOfTrillEnd((NChord *) elem, &dest_measure_start_time, &count_of_measures);
					if (ornaments_->pcount() < 1) {
						*ornaments_ << "mussym above " << staff_nr << ": ";
					}
					starttime = ((elem->midiTime_ - measure_start_time) / 3) / (double) (128 / curr_denom_);
					endtime   = ((tt - dest_measure_start_time) / 3) / (double) (128 / curr_denom_);
					if (count_of_measures) {
					 	timestring.sprintf("%dm + %f", count_of_measures, endtime + 1.0);
					}
					else {
						timestring.sprintf("%f", endtime + 1.0);
					}
					if (elem->trill_ > 0) {
						*ornaments_ << (starttime + 1.0) << " \"tr\" til " << timestring << ';';
					}
					else {
						*ornaments_ << (starttime + 1.0) << " \"lonelytrill\" til " << timestring << ';';
						bad = new badmeasure(ERR_LONELY_TRILL, staff_nr, bar_nr_, total / 3, countof128th_);
						badlist_.append(bad);
					}
				     }
				     if (elem->dynamic_) {
					dest_measure_start_time = measure_start_time;
					tt = voi->findTimeOfDynamicEnd(chord, &measure_start_time,
						 &dest_measure_start_time, &count_of_measures);
					if (tt < 0) {
						 *os_ << "damaged dynamic in staff " << staff_nr << " measure: " << bar_nr_ << '\0';
						KMessageBox::sorry(0, QString(os_->str()), kapp->makeStdCaption(i18n("???")));
					}
					else {
					   starttime = ((chord->midiTime_ - measure_start_time) / 3) / (double) (128 / curr_denom_);
					   endtime   = ((tt - dest_measure_start_time) / 3) / (double) (128 / curr_denom_);
					   if (count_of_measures) {
					   	timestring.sprintf("%dm + %f", count_of_measures, endtime + 1.0);
					   }
					   else {
						timestring.sprintf("%f", endtime + 1.0);
					   }
					   *dynamics_ << (chord->dynamicAlign_ ? '<' : '>') << "  below " << staff_nr << ": " <<
							(starttime + 1.0) << " til " << timestring << ';';
				        }
				     }
				     for (i = 0; i < NUM_LYRICS; i++) {
				     	lyrics = ((NChord *) elem)->getLyrics(i);
				     	if (lyrics) {
						(*lyricsLine_[i]) << lyrics2MUP(lyrics) << ' ';
						if (i+1 > lyricsLineCount) lyricsLineCount = i+1;
				     	}
				     }
				     break;
			case T_REST: 
				     if (loop1) {
					loop1 = false;
					out_ << staff_nr << " 1: ";
				     }
				     part = elem->getSubType(); 
				     if (part == MULTIREST) {
					bad = new badmeasure(ERR_MULTIREST, staff_nr, bar_nr_, total / 3, countof128th_);
					fatallist_.append(bad);
					if (divide_multi_rest(staff_nr, 1, ((NRest *) elem)->getMultiRestLength())) {
						total = 3*countof128th_;
						return SIMPLE_BAR;
					}
					break;
				     }

				     total += (elem->status_ & STAT_TRIPLET) ? 2 * part / 3 : part;
				     switch (elem->status_ & DOT_MASK) {
					case 1: total += part / 2; break;
					case 2: total += 3 * part / 4; break;
				     }
				     if ((elem->status_ & STAT_TRIPLET) && !intriplet) {
					intriplet = true;
					out_ << "{ ";
				     }
				     if (total > 3*countof128th_) {
					total = 0;
					bad = new badmeasure(ERR_NOTE_COUNT, staff_nr, bar_nr_, total / 3, countof128th_);
					badlist_.append(bad);
				     }
				     out_ << 3*128 / part;
				     switch (elem->status_ & DOT_MASK) {
					case 1:	out_ << "."; break;
					case 2:	out_ << ".."; break;
				     }
				     out_ << ((elem->status_ & STAT_HIDDEN) ? "s; " : "r; ");
				     if (elem->status_ & STAT_LAST_TRIPLET) {
					intriplet = false;
					out_ << " } above 3; ";
				     }
				     break;
			case T_SIGN: switch (elem->getSubType ()) {
					case SIMPLE_BAR: 
					case REPEAT_OPEN:
					case REPEAT_CLOSE:
					case REPEAT_OPEN_CLOSE:
							 if (loop1) {
								loop1 = false;
								out_ << staff_nr << " 1: ";
							 }
							 if (total != 3*countof128th_) {
								bad = new badmeasure(ERR_NOTE_COUNT, staff_nr, bar_nr_, total / 3, countof128th_);
								badlist_.append(bad);
							 }
							 subtype = elem->getSubType();
							 elem = voi->getNextPosition();
							 if (elem && elem->getType() == T_SIGN  && (elem->getSubType() & SPECIAL_ENDING)) {
								subtype |= elem->getSubType ();
								elem = voi->getNextPosition();
							 }
							 out_ << endl;
							if (lyricsLineCount) {
								if (lyricsLineCount == 1) {
									(*lyricsLine_[0]) << '\0';
									out_ << "lyrics " << staff_nr << ": \"" << lyricsLine_[0]->str() << "\";" << endl;
								}
								else {	
									for (i = 0; i < lyricsLineCount; i++) {
										(*lyricsLine_[i]) << '\0';
										out_ << "lyrics " << staff_nr << ":  [" << i+1 << "]\"" << lyricsLine_[i]->str() << "\";" << endl;
									}
								}
							}
							 return subtype;
					case TEMPO_SIGNATURE:
							sign = (NSign *) elem;
							if (loop1) {
								writeTempoSig(sign);
								pending_tempo_sig_ = 0;
							}
							else {
								bad = new badmeasure(ERR_TEMPO_PLACEMENT, staff_nr, bar_nr_, total / 3, countof128th_);
								fatallist_.append(bad);
								pending_tempo_sig_ = sign;
							}
							break;
					case VOLUME_SIG:
							sign = (NSign *) elem;
							if (loop1) {
								writeVolSig(staff_nr, sign);
								pending_elems_[staff_nr-1].pending_volsig = 0;
							}
							else {
								bad = new badmeasure(ERR_VOL_PLACEMENT, staff_nr, bar_nr_, total / 3, countof128th_);
								fatallist_.append(bad);
								pending_elems_[staff_nr-1].pending_volsig = sign;
							}
							break;
					case PROGRAM_CHANGE:
							sign = (NSign *) elem;
							if (loop1) {
								writeProgramChange(staff_nr, sign);
								pending_elems_[staff_nr-1].pending_program_change = 0;
							}
							else {
								bad = new badmeasure(ERR_PROG_PLACEMENT, staff_nr, bar_nr_, total / 3, countof128th_);
								fatallist_.append(bad);
								pending_elems_[staff_nr-1].pending_program_change = sign;
							}
							break;
				     }
				     break;
		}
		elem = voi->getNextPosition();
	}
	while (elem);
	out_ << endl;
	if (lyricsLineCount) {
		if (lyricsLineCount == 1) {
			(*lyricsLine_[0]) << '\0';
			out_ << "lyrics " << staff_nr << ": \"" << lyricsLine_[0]->str() << "\";" << endl;
		}
		else {	
			for (i = 0; i < lyricsLineCount; i++) {
				(*lyricsLine_[i]) << '\0';
				out_ << "lyrics " << staff_nr << ":  [" << i+1 << "]\"" << lyricsLine_[i]->str() << "\";" << endl;
			}
		}
	}
	return SIMPLE_BAR;
}

void NFileHandler::writeTempoSig(NSign *temposig) {
	out_ << "midi all: 0 \"tempo=" << temposig->getTempo() <<
		 "\";" << endl;
	out_ << "rom above all: 1 \"( \\(sm4n) = " << 
	temposig->getTempo() << " )\";" << endl;
}

void NFileHandler::writeVolSig(int staff_nr, NSign *volsig) {
	out_ << "midi " << staff_nr << " 1: 0 \"parameter=7," << volsig->getVolume() << "\";" << endl;
	out_ << "boldital above" << staff_nr << ": 1 \"";
	switch(volsig->getVolType()) {
		case V_PPPIANO : out_ << "ppp"; break;
		case V_PPIANO  : out_ << "pp"; break;
		case V_PIANO   : out_ << "p"; break;
		case V_FORTE   : out_ << "f"; break;
		case V_FFORTE  : out_ << "ff"; break;
		case V_FFFORTE : out_ << "fff"; break;
		default        : out_ << "mf"; break;
	}
	out_ << "\"; // volumesign" << endl;
}

void NFileHandler::writeProgramChange(int staff_nr, NSign *pChange) {
	out_ << "midi " << staff_nr << " 1: 0 \"program=" << pChange->getProgram() << "\";" << endl;
}

int NFileHandler::determineMultiRest(QList<NStaff> *stafflist) {
	NStaff *staff_elem;
	int len;

	staff_elem = stafflist->first();
	len = staff_elem->determineMultiRest();
	if (len == 0) return 0;
	for (staff_elem = stafflist->next(); staff_elem; staff_elem = stafflist->next()) {
		if (staff_elem->determineMultiRest() != len) return 0;
	}
	return len;
}

bool NFileHandler::divide_multi_rest(int staff_nr, int voice_nr, int multirestlength) {
	int part, len2, dotcount;
	int *iptr = pending_multi_rests_[staff_nr - 1] + (voice_nr - 1);

	if (*iptr == 0) {
		*iptr= multirestlength * 3 * countof128th_;
	}
	if (*iptr > 0) {
		if (*iptr >= 3*countof128th_) {
			part = 3*countof128th_;
		}
		else {
			part = *iptr;
		}
		*iptr -= part;
		while (part > 2) { 
			len2 = NVoice::quant(part, &dotcount, 3*128);
			part -= dotcount ? 3 * len2 / 2 : len2;
			out_ << 3*128 / len2;
			if (dotcount) out_ << '.';
			out_ << "r; ";
		}
	}
	if (*iptr != 0) out_ << endl;
	return (*iptr != 0);
}


	

	

void NFileHandler::writeVoiceElemsTill(int staff_nr, int voice_nr, NVoice *voi, int stopTime, int multirestlength) {
	NMusElement *elem;
	int total = 0;
	int part, tt;
	NNote *note, *note2;
	NChord *chord, *partner;
	bool inbeam = false;
	bool intriplet = false;
	bool something_written = false;
	int subtype;
	int dest_measure_start_time;
	int measure_start_time = -1;
	int i, dist;
	int count_of_measures;
	QString timestring;
	double starttime, endtime;
	bool loop1 = true, loop2;
	badmeasure *bad;
	NStaff *actual_staff;

	actual_staff = voi->getStaff();
	elem = voi->getCurrentPosition();
	if (multirestlength) {
		while(elem && elem->midiTime_ < stopTime) 
			elem = voi->getNextPosition();
		loop1 = false;
	}
	while (elem && elem->midiTime_ < stopTime) {
		switch (elem->getType()) {
			case T_CHORD: if (loop1) {
					loop1 = false;
					out_ << staff_nr << " " << voice_nr << ": ";
				     }
				     part = elem->getSubType(); 
				     total += (elem->status_ & STAT_TRIPLET) ? 2 * part / 3 : part;
				     switch (elem->status_ & DOT_MASK) {
					case 1: total += part / 2; break;
					case 2: total += 3 * part / 4; break;
				     }
				     if ((elem->status_ & STAT_TRIPLET) && !intriplet) {
					intriplet = true;
					out_ << "{ ";
				     }
				     if (total > 3*countof128th_) {
					total = 0;
					bad = new badmeasure(ERR_NOTE_COUNT, staff_nr, bar_nr_, total / 3, countof128th_);
					badlist_.append(bad);
				     }
				     chord = (NChord *) elem;

				     if (chord->status_ & STAT_SFORZ || chord->status_ & STAT_PORTA ||
				         chord->status_ & STAT_STPIZ || chord->status_ & STAT_SFZND ||
					 chord->status_ & STAT_STACC || chord->status_ & STAT_FERMT) {
				    	    out_ << "[with ";
				    	    if (chord->status_ & STAT_STACC)
						out_ << ".";
				    	    if (chord->status_ & STAT_SFORZ) 
					        out_ << "^";
					    if (chord->status_ & STAT_PORTA)
					        out_ << "-";
					    if (chord->status_ & STAT_STPIZ)
					        out_ << ",";
					    if (chord->status_ & STAT_SFZND)
					        out_ << ">";
					    if (chord->status_ & STAT_FERMT)
						out_ << "\"\\(ferm)\"";
					    out_ << "] ";
				     }
				     if (chord->getSubType() <= 3*64) {
					if (((chord->status_ & STAT_STEM_UP) &&  chord->getNoteList()->first()->line > 3)  ||
					    (!(chord->status_ & STAT_STEM_UP) && chord->getNoteList()->first()->line  < 4)) {
						out_ << ((chord->status_ & STAT_STEM_UP) ? "[up]" : "[down] ");
					}
				     }
				     out_ << 3*128 / part;
				     switch (elem->status_ & DOT_MASK) {
					case 1:	out_ << "."; break;
					case 2:	out_ << ".."; break;
				     }
				     loop2 = true;
			  	     for (note = elem->getNoteList()->first(); note; note = elem->getNoteList()->next()) {
					     pitchOut( note, &(actual_staff->actualClef_), true);
					     if (!drum_problem_written_  && (note->status & BODY_MASK)) {
						drum_problem_written_ = true;
						bad = new badmeasure(ERR_DRUM_STAFF, staff_nr, 0 /*dummy */, 3 /* dummy */, countof128th_);
						badlist_.append(bad);
					     }
					     if (loop2) {
						loop2 = false;
						if (chord->status_ & STAT_SLURED) {
							dist = voi->computeSlurDist(chord);
							if (dist > 1) {
								bad = new badmeasure(ERR_SLUR, staff_nr, bar_nr_, 0, 0);
								badlist_.append(bad);
								out_ << "<" << dist << ">";
							}
							else {
								partner = chord->getSlurPartner();
								note2 = partner->getNoteList()->first();
								out_ << "<" ;
					    			pitchOut( note2, &(actual_staff->actualClef_), false);
								out_ << ">" ;
							}
						}
					    }
				     }
				     if (elem->status_ & STAT_BEAMED) {
					if (!inbeam) {
						out_ << " bm";
						inbeam = true;
					 }
				     	 else if (elem->lastBeamed()) {
						out_ << " ebm";
						inbeam = false;
					 }
				     }
				     out_ << "; ";
				     if (elem->status_ & STAT_LAST_TRIPLET) {
					intriplet = false;
					out_ << " } above 3; ";
				     }
				     if (elem->trill_) {
					dest_measure_start_time = measure_start_time;
					tt = voi->findTimeOfTrillEnd((NChord *) elem, &dest_measure_start_time, &count_of_measures);
					if (ornaments_->pcount() < 1) {
						*ornaments_ << "mussym above " << staff_nr << ": ";
					}
					starttime = ((elem->midiTime_ - measure_start_time) / 3) / (double) (128 / curr_denom_);
					endtime   = ((tt - dest_measure_start_time) / 3) / (double) (128 / curr_denom_);
					if (count_of_measures) {
					 	timestring.sprintf("%dm + %f", count_of_measures, endtime + 1.0);
					}
					else {
						timestring.sprintf("%f", endtime + 1.0);
					}
					if (elem->trill_ > 0) {
						*ornaments_ << (starttime + 1.0) << " \"tr\" til " << timestring << ';';
					}
					else {
						*ornaments_ << (starttime + 1.0) << " \"lonelytrill\" til " << timestring << ';';
						bad = new badmeasure(ERR_LONELY_TRILL, staff_nr, bar_nr_, total / 3, countof128th_);
						badlist_.append(bad);
					}
				     }
				     something_written = true;
				     break;
			case T_REST: if (loop1) {
					loop1 = false;
					out_ << staff_nr << " " << voice_nr << ": ";
				     }
				     part = elem->getSubType(); 
				     total += (elem->status_ & STAT_TRIPLET) ? 2 * part / 3 : part;
				     switch (elem->status_ & DOT_MASK) {
					case 1: total += part / 2; break;
					case 2: total += 3 * part / 4; break;
				     }
				     if ((elem->status_ & STAT_TRIPLET) && !intriplet) {
					intriplet = true;
					out_ << "{ ";
				     }
				     if (total > 3*countof128th_) {
					total = 0;
					bad = new badmeasure(ERR_NOTE_COUNT, staff_nr, bar_nr_, total / 3, countof128th_);
					badlist_.append(bad);
				     }
				     out_ << 3*128 / part;
				     switch (elem->status_ & DOT_MASK) {
					case 1:	out_ << "."; break;
					case 2:	out_ << ".."; break;
				     }
				     out_ << ((elem->status_ & STAT_HIDDEN) ? "s; " : "r; ");
				     if (elem->status_ & STAT_LAST_TRIPLET) {
					intriplet = false;
					out_ << " } above 3; ";
				     }
				     something_written = true;
				     break;
		}
		elem = voi->getNextPosition();
	}
	if (something_written) out_ << endl;
}
	
bool NFileHandler::readStaffs(const char *fname, QList<NVoice> *voilist, QList<NStaff> *stafflist, NMainFrameWidget *mainWidget) {
	FILE *fp;
	extern int YYRESTART(FILE *fp);
	extern int YYPARSE();
	extern int YYLINENO;
	int parser_return;
	NVoice *voice_elem;
	NStaff *staff_elem;
	QList<NVoice> newVoices;
	QList<NStaff> newStaffs;

	parser_params.fname = fname;
	parser_params.mainWidget = mainWidget;
	parser_params.newStaffs = &newStaffs;
	parser_params.newVoices = &newVoices;

	if ((fp = fopen(fname, "r")) == NULL) {
		*os_ << "error opening file " << fname << '\0';
		KMessageBox::sorry
		  (0,
		   QString(os_->str()),
		   kapp->makeStdCaption(i18n("???")));
		return false;
	}
	YYRESTART(fp);
	YYLINENO = 1;
	init_parser_variables();
	parser_return = YYPARSE();
	fclose(fp);
	cleanup_parser_variables();
	if (parser_return != 0) {
		while (!newVoices.isEmpty()) {
			newVoices.first();
			newVoices.current()->emptyVoice();
			newVoices.remove();
		}
		return false;
	}
		
	while (!voilist->isEmpty()) {
		voilist->first();
		voilist->current()->emptyVoice();
		voilist->remove();
	}
	for (voice_elem = newVoices.first(); voice_elem; voice_elem = newVoices.next()) {
		voilist->append(voice_elem);
	}
	stafflist->clear();
	for (staff_elem = newStaffs.first(); staff_elem; staff_elem = newStaffs.next()) {
		stafflist->append(staff_elem);
	}
	mainWidget->scTitle_ = parser_params.scTitle_;
	mainWidget->scSubtitle_ = parser_params.scSubtitle_;
	mainWidget->scAuthor_ = parser_params.scAuthor_;
	mainWidget->scLastAuthor_ = parser_params.scLastAuthor_;
	mainWidget->scComment_ = parser_params.scComment_;
	mainWidget->setParamsEnabled(parser_params.enableParams);
	mainWidget->setSaveWidth(parser_params.paperwidth);
	mainWidget->setSaveHeight(parser_params.paperheight);
	mainWidget->setWithMeasureNums(parser_params.with_measnum);
	return true;
}

QString NFileHandler::lyrics2MUP(QString *lyrics) {
	QString ret;
	QRegExp reg;

	ret = QString(*lyrics);
	reg = QRegExp("\"");
	ret.replace (reg, "\\\"");
	reg = QRegExp("-");
	ret.replace (reg, "~");
	reg = QRegExp("~$");
	ret.replace (reg, "-");
	return ret;
}

